/*  File   : awk.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for AWK and Hawk.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 awk.c	1.6";
#endif/*lint*/

#include "many2h.h"
#include "language.h"

/*  We generally try to keep track of whether we are expecting
    an operator or an operand, so that we can tell the difference
    between Id'Attr and 'x' in Ada and the difference between
    x/y and x ~ /y/ in AWK.  There are no #defines in AWK to mess this up.
*/
static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    if (state == String_Open) {
	newcol = terminator_column(column, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = Term_Ender;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	    t = 0;
	}
    } else {
	t = 0;    
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter_or_underscore(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].length = column - token_row[t].start;
	    k = is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = Identifier;
		state = Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {
		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    newcol = column;
	    do column++;
		while (is_a_digit_or_underscore(line_buffer[column]));
	    c = line_buffer[column];
	    if (c == '.') {
		do column++;
		    while (is_a_digit_or_underscore(line_buffer[column]));
		if (dncase[line_buffer[column]] == 'e'
		 && !is_a_letter(line_buffer[column+1])
                ) {
		    column++;
		    c = line_buffer[column];
		    if (c == '+' || c == '-') column++;
		    while (is_a_digit_or_underscore(line_buffer[column]))
			column++;
		}
	    }
	    token_row[t].type   = Number;
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
	} else
	if (c == '#') {
	    token_row[t].type = Comment_Opener;
	    do column++; while (line_buffer[column] == c);
	    token_row[t].length = column - token_row[t].start;
	    t++;
	    while (line_buffer[column] == ' ') column++;
	    if (line_buffer[column] != '\0') {
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = line_length - column;
		t++;
	    }
	    token_row[t].type   = Comment_Closer;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    token_row[t].length = 0;
	    column = line_length;
	} else
	if (c == '`') {
	    lq_count++;
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (c == '\'' && lq_count > 0) {
	    lq_count--;
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == '"'
         || c == '\''
         || (c == '/' && state == Other_Token && opchar == '\0')
        ) {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
        if (c == opchar) {
	    token_row[t].type = Special;	/* it is an operator */
	    for (;;) {
	        column++;
	        if (line_buffer[column] == c) {
	            column++;
	            break;
		} else
		if (line_buffer[column] == '\0') {
		    where_am_i(token_row[t].start);
		    fprintf(stderr, "Missing closing %c\n", c);
		    break;
		}
	    }
	    token_row[t].length = column - token_row[t].start;
	    state = Other_Token;
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = c == '}' ? Other_Token : Term_Ender;
	} else
	if (c == ',' || c == ';') {
	    /* What about colon?  There are no labels in AWK. */
	    /* When used as part of the if?then:else construct, */
	    /* it is NOT a comma, it is an operator, and should */
	    /* have a space on either side, like other wide scope */
	    /* operators. */
	    token_row[t].type    = Comma;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
        if (c == '<' && opchar != '\0') {
	    token_row[t].type = Special;
	    if (line_buffer[column+1] == '<' && line_buffer[column+2] == '-') {
	        token_row[t].length = 3;
	    } else
	    if (line_buffer[column+1] == '-') {
	        token_row[t].length = 2;
	    } else {
	        token_row[t].length = 1;
	    }
	    column += token_row[t].length;
	    state = Other_Token;
	} else
	if (is_a_special_character(c)) {
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	} else {
	    /* @ plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
    if (state == Term_Ender) state = Other_Token;  /* EOL = ; in AWK */
}


static char *awk_keywords[] = {
    "BEGIN",
    "break",
    "continue",
    "do",
    "else",
    "END",
    "exit",
    "for",
    "func",
    "function",
    "getline",
    "if",
    "in",
    "next",
    "print",
    "printf",
    "return",
    "while",
    (char *)0
};

struct Language awk = {
    "AWK",
    ".awk.hwk.hawk.twk.tawk",
    "([{",
    ")]}",
    "_",
    "~!#$%^&*-+=|\\<>?:/.",
    '\\',
    '\0',
    Is_Case_Sensitive,
    awk_keywords,
    tokenise
};


/*  Hmm.  The S language, as supported by R, is pretty close to
    AWK except for not having /.../ regular expression literals.
*/

static char *R_keywords[] = {
    "break",
    "else",
    "FALSE",
    "for",
    "function",
    "if",
    "in",
    "Inf",
    "NA",
    "NaN",
    "next",
    "NULL",
    "repeat",
    "switch",		/* NOT a keyword but worth highlighting */
    "while",
    (char *)0
};

struct Language R = {
    "R",
    ".R.S",
    "([{",
    ")]}",
    "_.",
    "~!@#$^&*-+=|\\<>?:/.",
    '\\',
    '%',
    Is_Case_Sensitive,
    R_keywords,
    tokenise
};



