/*  File   : basic.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for BASIC-like languages, currently just OPL.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 basic.c	1.4";
#endif/*lint*/

#include "many2h.h"
#include "language.h"


static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;
    static int initialised = 0;

    if (!initialised) {
	/* We need to add a comment keyword list to all language */
	/* descriptions.  BASIC isn't the only language to have  */
	/* a keyword like REM. */
        add_kinded_keyword("rem", 3, -1);
        initialised = 1;
    }

    for (column = 0; line_buffer[column] == ' '; column++) {}
    /* BASIC strings cannot span multiple lines. */
    /* BASIC comments cannot span multiple lines. */
    t = 0;    
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    c = line_buffer[column];
	    if (c == '$' || c == '%' || c == '&' || c == '#') ++column;
	    /* <id>: is a procedure name; <id>:: is a label name */
	    if (line_buffer[column] == ':') column++;
	    if (line_buffer[column] == ':') column++;
	    token_row[t].length = column - token_row[t].start;
	    k = is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = Identifier;
		state = Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {
		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    do column++;
		while (is_a_digit_or_underscore(line_buffer[column]));
	    c = line_buffer[column];
	    if (c == '.' && line_buffer[column+1] != c) {
		do column++;
		    while (is_a_digit_or_underscore(line_buffer[column]));
	    }
	    if (dncase[line_buffer[column]] == 'e'
	     && !is_a_letter(line_buffer[column+1])
	    ) {
		column++;
		c = line_buffer[column];
		if (c == '+' || c == '-') column++;
		while (is_a_digit_or_underscore(line_buffer[column]))
		    column++;
	    }
	    /* x (+|-) y% = x * (1 (+|-) y/100) */
	    /* x (*|/) y% = x (*|/) (y/100) */
	    if (line_buffer[column] == '%') column++;
	    token_row[t].type   = Number;
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
	} else
	if (c == '$' || c == '&') {
	    /* $[:hex:]{1,4} is a short hexadecimal integer in OPL */
	    /* &[:hex:]{1,8} is a long  hexadecimal integer in OPL */
	    do column++;
	        while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].type   = Number;
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
        } else	    
	if (c == '%') {
	    /* <id>% is an integer variable. */
	    /* <num>% is a percentage operator. */
	    /* But %<char> is a character literal. */
	    token_row[t].type = Character_Name;
	    column++;
	    if (line_buffer[column] == '\0') {
		where_am_i(column);
		fprintf(stderr, "No character after %%\n");
		token_row[t].length = 1;
	    } else {
	        column++;
	        token_row[t].length = 2;
	    }
	} else
	if (c == '`') {
	    lq_count++;
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (c == '\'' && lq_count > 0) {
	    lq_count--;
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == '\'') {
	    token_row[t].type = Comment_Opener;
	    do column++; while (line_buffer[column] == c);
	    token_row[t].length = column - token_row[t].start;
	    t++;
	    while (line_buffer[column] == ' ') column++;
	    if (line_buffer[column] != '\0') {
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = line_length - column;
		t++;
	    }
	    token_row[t].type   = Comment_Closer;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    token_row[t].length = 0;
	    column = line_length;
	} else
	if (c == '"') {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, terminator[0]);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else{
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    }
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == ',' || c == ';' || c == ':') {
	    token_row[t].type    = Comma;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_special_character(c)) {
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	} else {
	    /* ~ ` ! # $ ? plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


static char *keywords[] = {
    "and", 
    "app",
    "below",			/* preprocessor */
    "break",
    "const",
    "continue",
    "declare",
    "do",
    "else",
    "elseif",
    "enda",
    "endfor",			/* preprocessor */
    "endif",
    "endp",
    "endv",
    "endwh",
    "external",
    "for",			/* preprocessor */
    "from",			/* preprocessor */
    "global",
    "goto",
    "if",
    "include",
    /*"input",*/		/* we don't really want this highlighted? */
    "local",
    "not",
    "or",
    "proc",
    /*"print",*/		/* we don't really want this highlighted? */
    "return",
    "to",			/* preprocessor */
    "until",
    "vector",
    "while",
};

struct Language opl = {
    "OPL",
    ".opl.op",
    "(",
    ")",
    "_",
    "@^<>+-*/=.",
    '\0',
    '\0',
    Not_Case_Sensitive,
    keywords,
    tokenise
};

