/*  File   : c.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for C, C++, and Java.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 c.c	1.24";
#endif/*lint*/

#include "many2h.h"
#include "language.h"


/*  We generally try to keep track of whether we are expecting
    an operator or an operand, so that we can tell the difference
    between Id'Attr and 'x' in Ada and the difference between
    x/y and x ~ /y/ in AWK.  C makes this extremely hard in two
    ways:  #define <id> <anything> means that it is actually quite
    common for a number, string, identifier &c to follow an identifier
    in that context, and (type)expr casts mean that it is common for
    things to follow a right parenthesis.  Since C has no such
    context-sensitive tokens, it seems best to simply omit the error
    reports rather than try to patch them up.  The context maintenance
    is kept, though, in order to keep the C and AWK tokenisers similar.
*/
static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    if (state == String_Open) {
	newcol = terminator_column(column, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = Term_Ender;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	    t = 0;
	}
    } else
    if (state == Comment_Open) {
	newcol = terminator_column(column, EOF);
	token_row[0].type   = Comment_Element;
	token_row[0].start  = column;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = Comment_Closer;
	token_row[1].start  = newcol;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = old_state;
	t = 2;
    } else {
	t = 0;    
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter_or_underscore(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].length = column - token_row[t].start;
	    k = (token_row[t].length > 2 &&
	         line_buffer[column-2] == '_' &&
	         line_buffer[column-1] == 't'
	        ) || is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = Identifier;
		state = t == 2 && token_row[0].text[0] == '#'
		      ? Other_Token : Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {
		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    newcol = column;
	    do column++;
		while (is_a_digit_or_underscore(line_buffer[column]));
	    c = line_buffer[column];
	    if (dncase[c] == 'x' && line_buffer[newcol] == '0'
             && column - newcol == 1
	    ) {
		do column++;
		    while (is_a_letter_or_digit(line_buffer[column]));
	    } else
	    if (c == '.') {
		do column++;
		    while (is_a_digit_or_underscore(line_buffer[column]));
		if (dncase[line_buffer[column]] == 'e'
		 && !is_a_letter(line_buffer[column+1])
                ) {
		    column++;
		    c = line_buffer[column];
		    if (c == '+' || c == '-') column++;
		    while (is_a_digit_or_underscore(line_buffer[column]))
			column++;
		}
	    }
	    while (is_a_letter(line_buffer[column])) column++;
	    token_row[t].type   = Number;
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
	} else
	if (c == '/' && line_buffer[column+1] == c) {
	    token_row[t].type = Comment_Opener;
	    column += 2;
	    while (line_buffer[column] == c) column++;
	    token_row[t].length = column - token_row[t].start;
	    t++;
	    while (line_buffer[column] == ' ') column++;
	    if (line_buffer[column] != '\0') {
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = line_length - column;
		t++;
	    }
	    token_row[t].type   = Comment_Closer;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    token_row[t].length = 0;
	    column = line_length;
	} else
	if (c == '/' && line_buffer[column+1] == '*') {
	    static unsigned char const ender[3] = {'*', '/', '\0'};

	    token_row[t].type = Comment_Opener;
	    token_row[t].length = 2;
	    column += 2;
	    while (line_buffer[column] == ' ') column++;
	    terminator = ender;
	    termlength = 2;
	    newcol = terminator_column(column, EOF);
	    if (line_buffer[column] != '\0' && newcol != column) {
		t++;
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		if (newcol >= 0) {
		    while (newcol > 0 && line_buffer[newcol-1] == ' ')
			newcol--;
		    token_row[t].length = newcol - column;
		    newcol = terminator_column(newcol, EOF);
		} else {
		    token_row[t].length = line_length - column;
		}
	    }
	    if (newcol >= column) {
		t++;
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[newcol];
		token_row[t].start  = newcol;
		token_row[t].length = termlength;
		column = newcol+termlength;
	    } else {
		old_state = state;
		state = Comment_Open;
		column = line_length;
	    }
	} else
	if (c == '`') {
	    lq_count++;
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (c == '\'' && lq_count > 0) {
	    lq_count--;
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == '"' || c == '\'') {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == ',' || c == ';') {
	    /* The colon ':' has two uses in C. */
	    /* When used after a label or case label, it IS a comma */
	    /* and should be spaced like one. */
	    /* When used as part of the if?then:else construct, */
	    /* it is NOT a comma, it is an operator, and should */
	    /* have a space on either side, like other wide scope */
	    /* operators. */
	    /* It is impossible for this program to tell which use */
	    /* is intended, especially given the preprocessor. */
	    token_row[t].type    = Comma;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_special_character(c)) {
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	} else {
	    /* @ $ plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


static char *c_keywords[] = {
    "asm",		/* A common extension */
    "assert",		/* Really a macro, but very important */
    "__attribute",	/* For GCC */
    "auto",
    "bool",		/* New in C9x */
    "_Bool",		/* New in C9x (stdbool.h expands bool to this) */
    "break",
    "case",
    "char",
    "complex",		/* New in C9x */
    "__complex__",	/* GCC version */
    "const",
    "continue",
    "default",
    "define",		/* cpp & m4 */
    "defined",		/* cpp */
    "div_t",
    "do",
    "double",
    "elif",		/* cpp */
    "else",
    "endif",		/* cpp */
    "enum",
    "error",		/* cpp */
    "extern",
    "false",		/* New in C9X */
    "far",		/* DOS-specific */
    "_far",		/* DOS-specific */
    "__far",		/* DOS-specific */
    "float",
    "for",
    "goto",
    "huge",		/* DOS-specific */
    "_huge",		/* DOS-specific */
    "__huge",		/* DOS-specific */
    "if",
    "ifdef",		/* cpp & m4 */
    "ifndef",		/* cpp */
    "imaginary",	/* New in C9X */
    "include",		/* cpp & m4 */
    "inline",		/* New in C9X */
    "__inline",		/* GCC version */
    "int",
    "jmp_buf",		/* Not really a keyword, but a standard type */
    "long",
    "near",		/* DOS-specific */
    "_near",		/* DOS-specific */
    "__near",		/* DOS-specific */
    "pragma",		/* cpp */
    "register",
    "restrict",		/* New in C9X */
    "return",
    "short",
    "signed",
    "sizeof",
    "static",
    "struct",
    "switch",
    "true",		/* New in C9X */
    "typedef",
    "undef",		/* cpp */
    "union",
    "unsigned",
    "void",
    "volatile",
    "while",
#if 0
    /* The following type names come from the standard.
       POSIX adds so many more that I decided to treat _every_
       identifier ending with _t as a keyword,
       so we don't need these any more.
    */
    "clock_t",
    "ldiv_t",
    "lldiv_t",
    "ptrdiff_t",
    "size_t",
    "time_t",
    "wchar_t",
#endif
    (char *)0
};

struct Language c89 = {
    "C",
    ".c.h.ic.ih.l.y",
    "([{",
    ")]}",
    "_$",
    "~!#%^&*-+=|\\<>?:/.",
    '\\',
    '\0',
    Is_Case_Sensitive,
    c_keywords,
    tokenise
};


/*  This list includes C++0x.  */

static char *cpp_keywords[] = {
    "alignof",		/* New in C++0x */
    "and",
    "and_eq",
    "asm",
    "assert",		/* Really a macro, but very important */
    "__attribute",	/* For GCC */
    "auto",
    "bitand",
    "bitor",
    "bool",		/* C++ & new in C9x */
    "break",
    "case",
    "catch",		/* C++ */
    "char",
    "char16_t",		/* New in C++0x */
    "char32_t",		/* New in C++0x */
    "class",		/* C++ */
    "clock_t",		/* standard type name */
    "compl",
    "complex",
    "const",
    "constexpr",	/* New in C++0x */
    "const_cast",
    "continue",
    "decltype",		/* New in C++0x */
    "default",
    "define",		/* cpp & m4 */
    "defined",		/* cpp */
    "delete",		/* C++ */
    "div_t",
    "do",
    "double",
    "dynamic_cast",
    "elif",		/* cpp */
    "else",
    "endif",		/* cpp */
    "enum",
    "error",		/* cpp */
    "explicit",
    "export",
    "extern",
    "false",		/* C++ & new in C9X */
    "float",
    "for",
    "friend",
    "goto",
    "if",
    "ifdef",		/* cpp & m4 */
    "ifndef",		/* cpp */
    "include",		/* cpp & m4 */
    "inline",		/* New in C9X */
    "int",
    "jmp_buf",		/* Not really a keyword, but a standard type */
    "ldiv_t",
    "lldiv_t",
    "long",
    "mutable",
    "namespace",	/* C++ */
    "new",		/* C++ */
    "noexcept",		/* New in C++0x */
    "not",
    "not_eq",
    "nullptr",		/* New in C++0x */
    "operator",
    "or",
    "or_eq",
    "pragma",		/* cpp */
    "private",		/* C++ */
    "protected",	/* C++ */
    "ptrdiff_t",	/* Standard type name */
    "public",		/* C++ */
    "register",
    "reinterpret_cast",
    "restrict",		/* New in C9X */
    "return",
    "short",
    "signed",
    "size_t",		/* standard type name */
    "sizeof",
    "static",
    "static_assert",	/* New in C++0x */
    "static_cast",
    "string",		/* Not a keyword but a type */
    "struct",
    "switch",
    "template",		/* C++ */
    "this",		/* C++ */
    "thread_local",	/* New in C++0x */
    "throw",		/* C++ */
    "time_t",		/* standard type name */
    "true",		/* C++ & new in C9X */
    "try",
    "typedef",
    "typeid",
    "typename",
    "undef",		/* cpp */
    "union",
    "unsigned",
    "using",
    "virtual",		/* C++ */
    "void",
    "volatile",
    "wchar_t",
    "while",
    "xor",
    "xor_eq",
    (char *)0
};

struct Language cpp = {
    "C++",
    ".C.cc.cxx.cpp.H.hh.hxx.hpp",
    "([{",
    ")]}",
    "_",
    "~!#%^&*-+=|\\<>?:/.",
    '\\',
    '\0',
    Is_Case_Sensitive,
    cpp_keywords,
    tokenise
};


static char *nqc_keywords[] = {
    "__nolist",		/* not in my NQC */
    "abs",
    "acquire",
    "asm",
    "break",
    "case",
    "catch",
    "const",
    "continue",
    "default",
    "define",		/* cpp & m4 */
    "defined",		/* cpp */
    "do",
    "elif",		/* cpp */
    "else",
    "endif",		/* cpp */
    "error",		/* cpp */
    "false",		/* New in C9X */
    "for",
    "goto",
    "if",
    "ifdef",		/* cpp & m4 */
    "ifndef",		/* cpp */
    "include",		/* cpp & m4 */
    "int",
    "monitor",
    "repeat",
    "return",
    "sign",
    "start",
    "stop",
    "sub",
    "switch",
    "task",
    "true",		/* New in C9X */
    "undef",		/* cpp */
    "void",
    "while",
    (char *)0
};

struct Language nqc = {
    "NQC",
    ".nqc.nqh",
    "([{",
    ")]}",
    "_$",
    "~!#%^&*-+=|\\<>?:/.@",		/* @ has been replaced by asm */
    '\\',
    '\0',
    Is_Case_Sensitive,
    nqc_keywords,
    tokenise
};


static char *java_keywords[] = {
    "Array",
    "ArrayList",
    "BigDecimal",
    "BigInteger",
    "Bitset",
    "Boolean",
    "Byte",
    "Calendar",
    "Character",
    "Class",
    "ClassLoader",
    "Cloneable",
    "Constructor",
    "Date",
    "Dictionary",
    "Double",
    "Enumeration",
    "Field",
    "Float",
    "HashMap",
    "Hashtable",
    "Integer",
    "Locale",
    "Long",
    "Math",
    "Member",
    "Method",
    "Modifier",
    "Number",
    "Object",
    "Observable",
    "Observer",
    "Process",
    "Properties",
    "Random",
    "Runnable",
    "Short",
    "Stack",
    "String",
    "StringBuffer",
    "StringTokenizer",
    "System",
    "Thread",
    "Vector",
    "Void",

    "abstract",
    "assert",		/* keyword in Java 1.4 */
    "boolean",
    "break",
    "byte",
    "case",
    "catch",
    "char",
    "class",
    "continue",
    "default",
    "do",
    "double",
    "else",
    "enum",		/* keyword in Java 1.5 */
    "extends",
    "false",
    "final",
    "finally",
    "float",
    "for",
    "goto",		/* Not used yet, but is a reserved word */
    "if",
    "implements",
    "import",
    "instanceof",
    "int",
    "interface",
    "long",
    "new",
    "null",
    "package",
    "private",
    "protected",
    "public",
    "return",
    "short",
    "static",
    "super",
    "switch",
    "synchronized",
    "this",
    "throw",
    "throws",
    "transient",
    "true",
    "try",
    "void",
    "volatile",
    "while",
    (char *)0
};

struct Language java = {
    "Java",
    ".java.jav.jva",
    "([{",
    ")]}",
    "_$",
    "@~!#%^&*-+=|\\<>?:/.",
    '\\',
    '\0',
    Is_Case_Sensitive,
    java_keywords,
    tokenise
};

static char *d_keywords[] = {
    "assert",
    "bool",
    "break",
    "case",
    "clock_t",		/* standard type name */
    "char",
    "compl",
    "const",
    "continue",
    "define",		/* cpp & m4 */
    "defined",		/* cpp */
    "div_t",
    "do",
    "elif",		/* d & cpp */
    "else",
    "endif",		/* cpp */
    "enum",
    "error",		/* cpp */
    "esac",
    "extern",
    "false",		/* New in C9X */
    "fi",
    "for",
    "if",
    "ifdef",		/* cpp & m4 */
    "ifndef",		/* cpp */
    "include",		/* cpp & m4 */
    "int",
    "jmp_buf",		/* Not really a keyword, but a standard type */
    "ldiv_t",
    "lldiv_t",
    "long",
    "od",
    "ptrdiff_t",	/* Standard type name */
    "real",
    "return",
    "short",
    "size_t",		/* standard type name */
    "struct",
    "then",
    "time_t",		/* standard type name */
    "true",		/* New in C9X */
    "type",
    "undef",		/* cpp */
    "union",
    "unsigned",
    "until",
    "void",
    "volatile",
    "while",
    (char *)0
};

struct Language d04 = {
    "D",
    ".d.dh",
    "([{",
    ")]}",
    "_$",
    "~!#%^&*-+=|\\<>@:/.",
    '\\',
    '\0',
    Is_Case_Sensitive,
    d_keywords,
    tokenise
};


static char *rpcl_keywords[] = {
    "bool",
    "case",
    "const",
    "default",
    "double",
    "enum",
    "float",
    "hyper",
    "int",
    "opaque",
    "program",
    "quadruple",
    "string",
    "struct",
    "switch",
    "typedef",
    "union",
    "unsigned",
    "version",
    "void",
    (char *)0
};

/* I am not sure how to handle the "%" prefix used by rpcgen.
   If a line begins with "%" the rest of the line is C, not
   rpcl.  We have the same kind of problem with Lex and Yacc.
   Code switching might use
    - whole line prefixes like rpcgen	
    - contextual line processing (like 'after the second %%' in lex)
    - special open/close brackets ('%{' '%}')
    - ordinary brace counting (Yacc)
    - other magic ("C: '...c code...'" in my Smalltalk, similar in Mercury).
   The other language switched to is usually C but could be Fortran or
   Java or SML or Haskell.

   Dealing with code switching is a huge topic I'm not ready to deal
   with.  It will require a major rewrite.  It would be a good thing, mind!
   Right now, we DON'T deal with it, and the body of a % line will be
   processed as if it were rpcl, not C.  In particular, C keywords that
   are not rpcl keywords will NOT be recognised.
*/
struct Language rpcl = {
    "RPCL",			/* name */
    ".x",			/* suffixes */
    "([{",			/* opening brackets */
    ")]}",			/* closing brackets */
    "_$",			/* non-alphameric characters in identifiers */
    "~!#%^&*-+=|\\<>?:/.",	/* special characters */
    '\\',			/* escape for strings/characters */
    '\0',
    Is_Case_Sensitive,		/* are keywords case sensitive? */
    rpcl_keywords,		/* what are the keywords */
    tokenise			/* tokenising procedure */
};

