/*  File   : eiffel.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for Eiffel.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 eiffel.c	1.12";
#endif/*lint*/

#include "many2h.h"
#include "language.h"

static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    if (state == String_Open) {
	k = 0;			/* cols to skip when looking for " */
	if (termlength == 1) {
	    if (line_buffer[column] == escaper) {
		k = 1;		/* skip the leading % */
	    } else {
		where_am_i(column);
		fprintf(stderr, "Missing %c at start of string continuation\n",
			escaper);
	    }
	}
	newcol = terminator_column(column+k, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = Term_Ender;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	    t = 0;
	}
    } else
    if (state == Comment_Open) {
	newcol = terminator_column(column, EOF);
	token_row[0].type   = Comment_Element;
	token_row[0].start  = column;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = Comment_Closer;
	token_row[1].start  = newcol;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = old_state;
	t = 2;
    } else {
	t = 0;    
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].length = column - token_row[t].start;
	    k = is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = Identifier;
		state = Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {
		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    if (state != Other_Token) {
		where_am_i(column);
		fprintf(stderr, "number after term.\n");
	    }
	    newcol = column;
	    do column++;
		while (is_a_digit_or_underscore(line_buffer[column]));
	    c = line_buffer[column];
	    if (c == '.') {
		do column++;
		    while (is_a_digit_or_underscore(line_buffer[column]));
	    }
	    if (dncase[line_buffer[column]] == 'e'
	     && !is_a_letter(line_buffer[column+1])
            ) {
		column++;
		c = line_buffer[column];
		if (c == '+' || c == '-') column++;
		while (is_a_digit_or_underscore(line_buffer[column]))
		    column++;
	    }
	    token_row[t].type   = Number;
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
	} else
	if (c == '-' && line_buffer[column+1] == c) {
	    token_row[t].type = Comment_Opener;
	    column += 2;
	    while (line_buffer[column] == c) column++;
	    token_row[t].length = column - token_row[t].start;
	    t++;
	    while (line_buffer[column] == ' ') column++;
	    if (line_buffer[column] != '\0') {
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = line_length - column;
		t++;
	    }
	    token_row[t].type   = Comment_Closer;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    token_row[t].length = 0;
	    column = line_length;
	} else
	if (c == '`') {
	    lq_count++;
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (c == '\'' && lq_count > 0) {
	    lq_count--;
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == '"' || c == '\'') {
	    static unsigned char ender[3];

	    if (state != Other_Token) {
		where_am_i(column);
		fprintf(stderr, "Text literal after term.\n");
	    }
	    if (line_buffer[column+1] == '['
	     && line_buffer[column+2] == '\0'
	    ) {
		ender[0] = ']', ender[1] = c, ender[2] = '\0';
		termlength = 2;
		token_row[t].length = 2;
		column += 2;
	    } else {
		ender[0] = c, ender[1] = '\0';
		termlength = 1;
		token_row[t].length = 1;
		column++;
	    }
	    token_row[t].type   = String_Opener;
	    t++;
	    terminator = ender;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else
	    if (newcol < -1 && termlength == 1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (is_a_left_bracket(c)
         || (c == '<' && line_buffer[column+1] == c)
	) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = c == '<' ? 2 : 1;
	    column += token_row[t].length;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)
         || (c == '>' && line_buffer[column+1] == c)
	) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = c == '>' ? 2 : 1;
	    column += token_row[t].length;
	    state = Term_Ender;
	} else
	if (c == ':' && line_buffer[column+1] == '=') {
	    token_row[t].type    = Special;
	    token_row[t].length  = 2;
	    column += 2;
	    state = Other_Token;
	} else
	if (c == ',' || c == ';' || c == ':') {
	    token_row[t].type    = Comma;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_special_character(c)) {
	    token_row[t].type   = Special;
	    while (is_a_special_character(line_buffer[column])) column++;
	    token_row[t].length = column - token_row[t].start;
	    state = Other_Token;
	} else {
	    /* ` $ plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


static char *keywords[] = {
    "alias",
    "all",
    "and",
    "as",
    "check",
    "class",
    "create",
    "creation",
    "debug",
    "deferred",
    "do",
    "else",
    "elseif",
    "end",
    "ensure",
    "expanded",
    "export",
    "external",
    "false",
    "feature",
    "from",
    "frozen",
    "if",
    "implies",
    "indexing",
    "infix",
    "inherit",
    "insert",
    "inspect",
    "invariant",
    "is",
    "like",
    "local",
    "loop",
    "not",
    "obsolete",
    "old",
    "once",
    "or",
    "prefix",
    "redefine",
    "reference",
    "rename",
    "require",
    "rescue",
    "retry",
    "select",
    "separate",
    "strip",
    "then",
    "true",
    "undefine",
    "unique",
    "until",
    "variant",
    "when",
    "xor",
    (char *)0
};

struct Language eiffel = {
    "Eiffel",
    ".e.es.eif",
    "([{",
    ")]}",
    "_",
    "$&#<>+-*/=|.\\?!^~@",
    '%',
    '\0',   /* user-defined operators are done another way */
    Not_Case_Sensitive,
    keywords,
    tokenise
};

