/*  Header : formatter.h
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Specify Formatter record.
*/

#ifndef FORMATTER_H_
#define FORMATTER_H_ 19990623

#include <stddef.h>
#include "toktype.h"

enum Text_Style {
    Plain  =  00,
    Italic =  01,
    Bold   =  02,
    Ulined =  04,
    Typed  = 010,
    Allset = 017
};

struct Formatter {
    char const* const* start_style;
    char const* const* end_style;
    char const* const* char_map;
    void (*begin_output)(char const*, char const*, char const*, char const*);
    void (*end_output)(void);
    void (*begin_section)(char const *);
    void (*end_section)(void);
    void (*tab_from_to)(int, int);
    void (*end_token)(enum Token_Type);
    void (*finish_line)(void);
};

extern struct Formatter html;	/* HTML black-and-white highlights */
extern struct Formatter chtml;	/* HTML colour highlights */
extern struct Formatter roff;	/* Troff or nroff */
extern struct Formatter latex;	/* LaTeX (not TeX) */
extern struct Formatter plain;	/* plain text */
extern struct Formatter ansi;	/* ANSI black-and-white highlights */
extern struct Formatter colour;	/* ANSI colour highlights */
extern struct Formatter one_per_line;

struct Formatter const *opt_fmt(char const *);

extern void no_end_token(enum Token_Type);
extern void put_spaces(int);
#define     put_string(s) put_plain((unsigned char const *)(s))
extern void put_plain(unsigned char const *);
extern void put_mapped(unsigned char const *, struct Formatter const *const);
extern void put_styled(unsigned char const *, struct Formatter const *const,
                       enum Text_Style);
extern void write_chunk(unsigned char const *, size_t,
		        struct Formatter const *const, enum Text_Style);
extern void flush_output(void);

#endif
