/*  File   : latex.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Output table for generating LaTeX.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 latex.c	1.11";
#endif/*lint*/

#include "formatter.h"


static const char unknown[] = "?`{}";

static char const* const Start_Style[] = {
    "",
    "{\\it ",
    "{\\bf ",
    "{\\bf{\\it ",
    "{",
    "{{\\it ",
    "{{\\bf ",
    "{{\\bf{\\it ",
    "{\\tt ",
    "{\\tt{\\it ",
    "{\\tt{\\bf ",
    "{\\tt{\\bf{\\it ",
    "{\\tt{",
    "{\\tt{{\\it ",
    "{\\tt{{\\bf ",
    "{\\tt{{\\bf{\\it "
};

static char const *const End_Style[] = {
    "",
    "}",
    "}",
    "}}",
    "}",
    "}}",
    "}}",
    "}}}",
    "}",
    "}}",
    "}}",
    "}}}",
    "}}",
    "}}}",
    "}}}",
    "}}}}"
};

static char const *const Char_Map[] = {
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    "\\verb| |",          /*   */
    "!{}",        /* !; {} to suppress !` */
    "\\verb|\"|", /* " */
    "\\#",        /* # */
    "\\$",        /* $ */
    "\\%",        /* % */
    "\\&",        /* & */
    "\\verb|\'|", /* ' */
    "(",          /* ( */
    ")",          /* ) */
    "*",          /* * */
    "+",          /* + */
    ",",          /* , */
    "-",          /* - */
    ".",          /* . */
    "/",          /* / */
    "0",          /* 0 */
    "1",          /* 1 */
    "2",          /* 2 */
    "3",          /* 3 */
    "4",          /* 4 */
    "5",          /* 5 */
    "6",          /* 6 */
    "7",          /* 7 */
    "8",          /* 8 */
    "9",          /* 9 */
    ":",          /* : */
    ";",          /* ; */
    "$<$",        /* < */
    "=",          /* = */
    "$>$",        /* > */
    "?{}",        /* ?; {} to suppress ?` */
    "@",          /* @ */
    "A",          /* A */
    "B",          /* B */
    "C",          /* C */
    "D",          /* D */
    "E",          /* E */
    "F",          /* F */
    "G",          /* G */
    "H",          /* H */
    "I",          /* I */
    "J",          /* J */
    "K",          /* K */
    "L",          /* L */
    "M",          /* M */
    "N",          /* N */
    "O",          /* O */
    "P",          /* P */
    "Q",          /* Q */
    "R",          /* R */
    "S",          /* S */
    "T",          /* T */
    "U",          /* U */
    "V",          /* V */
    "W",          /* W */
    "X",          /* X */
    "Y",          /* Y */
    "Z",          /* Z */
    "[",          /* [ */
    "\\verb|\\|",   /* \ */
    "]",          /* ] */
    "\\verb|^|",  /* ^ */
    "\\_",        /* _ */
    "`",          /* ` */
    "a",          /* a */
    "b",          /* b */
    "c",          /* c */
    "d",          /* d */
    "e",          /* e */
    "f",          /* f */
    "g",          /* g */
    "h",          /* h */
    "i",          /* i */
    "j",          /* j */
    "k",          /* k */
    "l",          /* l */
    "m",          /* m */
    "n",          /* n */
    "o",          /* o */
    "p",          /* p */
    "q",          /* q */
    "r",          /* r */
    "s",          /* s */
    "t",          /* t */
    "u",          /* u */
    "v",          /* v */
    "w",          /* w */
    "x",          /* x */
    "y",          /* y */
    "z",          /* z */
    "\\{",         /* { */
    "$|$",         /* | */
    "\\}",         /* } */
    "\\verb|~|",   /* ~ */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    unknown,                 /* CONTROL CHARACTER */
    "~",                     /* no-break space */
    "!`",                    /* inverted exclamation mark */
    "/\\kern-.5em{c}",       /* cent sign */
    "{\\pounds}",            /* pound sterling sign */
    unknown,                 /* general currency sign */
    "=\\kern-.7em{Y}",       /* yen sign */
    "$\\ddagger$",           /* broken (vertical) bar [WRONG] */
    "{\\S}",                 /* section sign */
    "$\\ddot{\\verb| |}$",   /* umlaut (dieresis) */
    "{\\copyright}",         /* copyright sign */
    "${}^{\\underline{a}}$", /* ordinal indicator, feminine */
    "{\\footnotesize$\\ll$}",/* angle quotation mark, left */
    "$\\neg$",               /* not sign */
    "-",                     /* soft hyphen */
    "(R)",                   /* registered sign */
    "$\\bar{\\verb| |}$",    /* macron */
    "${}^\\circ$",           /* degree sign */
    "$\\pm$",                /* plus-or-minus sign */
    "${}^2$",                /* superscript two */
    "${}^3$",                /* superscript three */
    "$\\acute{\\verb| |}$",  /* acute accent */
    "$\\mu$",                /* micro sign */
    "{\\P}",                 /* pilcrow (paragraph sign) */
    "$\\cdot$",              /* middle dot */
    "$_{,}$",                /* cedilla; close but no cigar */
    "${}^1$",                /* superscript one */
    "${}^{\\underline{o}}$", /* ordinal indicator, masculine */
    "{\\footnotesize$\\gg$}",/* angle quotation mark, right */
    "$\\frac{1}{4}$",        /* fraction one-quarter */
    "$\\frac{1}{2}$",        /* fraction one-half */
    "$\\frac{3}{4}$",        /* fraction three-quarters */
    "?`",                    /* inverted question mark */
    "\\`{A}",                /* capital A, grave accent */
    "\\'{A}",                /* capital A, acute accent */
    "\\^{A}",                /* capital A, circumflex accent */
    "\\~{A}",                /* capital A, tilde */
    "\\\"{A}",               /* capital A, dieresis or umlaut mark */
    "{\\AA}",                /* capital A, ring */
    "{\\AE}",                /* capital AE diphthong (ligature) */
    "\\c{C}",                /* capital C, cedilla */
    "\\`{E}",                /* capital E, grave accent */
    "\\'{E}",                /* capital E, acute accent */
    "\\^{E}",                /* capital E, circumflex accent */
    "\\\"{E}",               /* capital E, dieresis or umlaut mark */
    "\\`{I}",                /* capital I, grave accent */
    "\\'{I}",                /* capital I, acute accent */
    "\\^{I}",                /* capital I, circumflex accent */
    "\\\"{I}",               /* capital I, dieresis or umlaut mark */
    "{Dh}",                  /* capital Eth, Icelandic; WRONG */
    "\\~{N}",                /* capital N, tilde */
    "\\`{O}",                /* capital O, grave accent */
    "\\'{O}",                /* capital O, acute accent */
    "\\^{O}",                /* capital O, circumflex accent */
    "\\~{O}",                /* capital O, tilde */
    "\\\"{O}",               /* capital O, dieresis or umlaut mark */
    "$\\times$",             /* multiply sign */
    "{\\O}",                 /* capital O, slash */
    "\\`{U}",                /* capital U, grave accent */
    "\\'{U}",                /* capital U, acute accent */
    "\\^{U}",                /* capital U, circumflex accent */
    "\\\"{U}",               /* capital U, dieresis or umlaut mark */
    "\\'{Y}",                /* capital Y, acute accent */
    "{Th}",                  /* capital THORN, Icelandic; WRONG */
    "{\\ss}",                /* small sharp s, German (sz ligature) */
    "\\`{a}",	             /* small a, grave accent */
    "\\'{a}",                /* small a, acute accent */
    "\\^{a}",                /* small a, circumflex accent */
    "\\~{a}",                /* small a, tilde */
    "\\\"{a}",               /* small a, dieresis or umlaut mark */
    "{\\aa}",                /* small a, ring */
    "{\\ae}",                /* small ae diphthong (ligature) */
    "\\c{c}",                /* small c, cedilla */
    "\\`{e}",      /* small e, grave accent */
    "\\'{e}",      /* small e, acute accent */
    "\\^{e}",      /* small e, circumflex accent */
    "\\\"{e}",    /* small e, dieresis or umlaut mark */
    "\\`{\\i}",      /* small i, grave accent */
    "\\'{\\i}",      /* small i, acute accent */
    "\\^{\\i}",      /* small i, circumflex accent */
    "\\\"{\\i}",    /* small i, dieresis or umlaut mark */
    "{dh}",       /* small eth, Icelandic; WRONG */
    "\\~{n}",      /* small n, tilde */
    "\\`{o}",      /* small o, grave accent */
    "\\'{o}",      /* small o, acute accent */
    "\\^{o}",      /* small o, circumflex accent */
    "\\~{o}",      /* small o, tilde */
    "\\\"{o}",    /* small o, dieresis or umlaut mark */
    "$\\div$",       /* divide sign */
    "{\\o}",       /* small o, slash */
    "\\`{u}",      /* small u, grave accent */
    "\\'{u}",      /* small u, acute accent */
    "\\^{u}",      /* small u, circumflex accent */
    "\\\"{u}",    /* small u, dieresis or umlaut mark */
    "\\'{y}",      /* small y, acute accent */
    "{th}",       /* small thorn, Icelandic; WRONG */
    "\\\"{y}",    /* small y, dieresis or umlaut mark */
};


#define write_latex_string(s) put_mapped((unsigned char const *)(s), &latex)

static void show(char const *tag, char const *val1, char const *val2) {
    put_string("\\");
    put_string(tag);
    put_string("{");
    write_latex_string(val1);
    if (val2 != 0 && *val2 != 0) write_latex_string(val2);
    put_string("}\n");
}


static void begin_output(
    char const *author,
    char const *date,
    char const *title1,
    char const *title2
) {
    put_string("\\documentclass[a4paper]{article}\n");
    if (title1 != 0 && *title1 != 0) show("title", title1, title2);
    if (author != 0 && *author != 0) show("author", author, "");
    if (date   != 0 && *date   != 0) show("date", date, "");
    put_string("\\begin{document}\n");
}


static void end_output(void) {
    put_string("\\end{document}\n");
}


static void begin_section(char const *title) {
    if (title != 0 && *title != 0) show("section*", title, "");
    put_string("\\noindent\n");
}


static void end_section(void) {
}


static void finish_line(void) {
    put_string("\\\\\\relax\n");
}


static void tab_from_to(int from, int to) {
    if (to > from) {
	put_string("\\verb|");
	put_spaces(to-from);
	put_string("|");
    }
}


struct Formatter latex = {
    Start_Style,
    End_Style,
    Char_Map,
    begin_output,
    end_output,
    begin_section,
    end_section,
    tab_from_to,
    no_end_token,
    finish_line
};

