/*  Header : many2h.h
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Types and variables used by all tokenisers.
*/

#include <ctype.h>
#include <limits.h>
#include <locale.h>
#include <stdio.h>
#include "toktype.h"

extern void where_am_i(int);

struct Token {
    unsigned char const *text;
    enum Token_Type      type;
    size_t               start;
    size_t               length;
};

extern struct Token  token_row[];

extern unsigned char line_buffer[];
extern size_t        line_length;
extern int           tab_width;
extern int read_line(FILE *);

#define LOWER   0x01
#define UPPER   0x02
#define LETTER  0x03
#define	DIGIT	0x04
#define	UNDERS	0x08
#define SPECLS  0x10
#define LBRACK  0x20                    /* single-char left bracket */
#define RBRACK  0x40			/* single-char right bracket */
#define is_lower(c)			(ctype[c]&LOWER)
#define is_upper(c)			(ctype[c]&UPPER)
#define is_a_letter(c)	 		(ctype[c]&LETTER)
#define is_a_letter_or_underscore(c)    (ctype[c]&(LETTER|UNDERS))
#define is_a_digit(c)  			(ctype[c]&DIGIT)
#define is_a_letter_or_digit(c)		(ctype[c]&(LETTER|DIGIT|UNDERS))
#define is_a_digit_or_underscore(c)	(ctype[c]&(DIGIT|UNDERS))
#define is_a_special_character(c)	(ctype[c]&SPECLS)
#define is_a_left_bracket(c)            (ctype[c]&LBRACK)
#define is_a_right_bracket(c)           (ctype[c]&RBRACK)
extern char ctype[256];

extern unsigned char tocase[3][1 + (int)UCHAR_MAX];
#define nocase (tocase[0])
#define dncase (tocase[1])
#define upcase (tocase[2])

extern char escaper;
extern char opchar;

/*  Between lines, the tokeniser can be in the following states:
    - the last token was one that can end a term
    - the last token was a tick
    - in the middle of a string (looking for a specified terminator)
    - in the middle of a comment (looking for a specified terminator)
    - anything else.
*/
enum Parser_State {
    Term_Ender,
    Prime,
    String_Open,
    Comment_Open,
    Other_Token,
    Extra_State		/* for SGML */
};
extern enum Parser_State     old_state;
extern enum Parser_State     state;
extern int                   lq_count;
extern unsigned char const * terminator;
extern size_t                termlength;

extern int terminator_column(int, int);

extern int  is_keyword(struct Token const *);
extern void add_user_keywords(char const *);
