/*  File   : reader.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Read lines.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 reader.c	1.7";
#endif/*lint*/

#include <stdlib.h>
#include <stdio.h>
#include "many2h.h"

#define Max_Line_Length (8192-100)

unsigned char line_buffer[Max_Line_Length + 100];
size_t        line_length = 0;
struct Token  token_row[Max_Line_Length];
int           tab_width = 8;

#define TAB   9
#define	LF   10
#define FF   12
#define CR   13
#define CTLZ 26
#define NBSP 160


/*  When reading the source data:
    - FF characters at the beginning of a line should eventually turn into
      horizontal rules, but are currently ignored.
    - FF characters elsewhere force a line break.
    - CR LF is a single line terminator.
    - CR without LF is a line terminator.
    - LF without CR is a line terminator.
    - TAB characters are expanded.
    - ^Z characters are reported as end of file.
    - other characters in the range 0..31, 127..159 are reported as errors
      that stop the program.
    - NBSP is turned into a space.
    - if the line expands to more than Max_Line_Length characters, that is
      reported as an error that stops the program.
*/
int read_line(register FILE *stream) {
    int column;
    int c;

    c = getc(stream);
    while (c == FF) {
	/* no action yet */
	c = getc(stream);
    }
    if (c == CTLZ || c == EOF) return 0;
    column = 0;
    for (;;) {
	if (c < ' ') {
	    if (c == TAB) {
		do line_buffer[column++] = ' '; while (column%tab_width != 0);
	    } else
	    if (c == LF) {
		break;
	    } else
	    if (c == CR) {
		c = getc(stream);
		if (c != LF) ungetc(c, stream);
		break;
	    } else
	    if (c == FF || c == CTLZ || c == EOF) {
		ungetc(c, stream);
		break;
	    } else {
		where_am_i(column+1);
		fprintf(stderr, "Not a printing character: %d\n", c);
		exit(EXIT_FAILURE);
	    }
	} else	
	if (c < 127 || c > NBSP) {
	    line_buffer[column++] = c;
	} else
	if (c == NBSP) {
	    line_buffer[column++] = ' ';
	} else {
	    where_am_i(column+1);
	    fprintf(stderr, "Not a printing character: %d\n", c);
	    exit(EXIT_FAILURE);
	}
	if (column >= Max_Line_Length) {
	    where_am_i(column+1);
	    fprintf(stderr, "Input line too long\n");
	    exit(EXIT_FAILURE);
	}
	c = getc(stream);
    }
    while (column > 0 && line_buffer[column-1] == ' ') column--;
    line_buffer[column] = '\0';
    line_length = column;
    return 1;
}

