/*  File   : scm.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for Scheme
    BEWARE : This unifies numbers and symbols.
*/

#include "many2h.h"
#include "language.h"

#define is_letter(c) ((unsigned)(((c)|32)-'a') < (unsigned)26)

static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    t = 0;
    if (state == String_Open) {
        if (line_buffer[column] == escaper) column++;
	newcol = terminator_column(column, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = Term_Ender;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	}
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter_or_underscore(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].length = column - token_row[t].start;
	    k = is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = is_a_letter(c) ? Identifier : Attribute;
		state = Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {
		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
        if (c == ';') {
            token_row[t].type = Comment_Opener;
            do column++; while (line_buffer[column] == c);
            token_row[t].length = column - token_row[t].start;
            t++;
            while (line_buffer[column] == ' ') column++;
            if (line_buffer[column] != '\0') {
                token_row[t].type   = Comment_Element;
                token_row[t].text   = &line_buffer[column];
                token_row[t].start  = column;
                token_row[t].length = line_length - column;
                t++;
            }
            token_row[t].type   = Comment_Closer;
            token_row[t].text   = &line_buffer[column];
            token_row[t].start  = column;
            token_row[t].length = 0;
            column = line_length;
        } else
	if (c == '"' || c == '|') {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (is_a_special_character(c)) {
	    token_row[t].type   = Special;
	    if (c == ',' && line_buffer[column+1] == '@') {
		token_row[t].length = 2;
		column += 2;
	    } else {
	        token_row[t].length = 1;
	        column += 1;
	    }
	    state = Other_Token;
        } else
        if (c == '#') {
	    column++;
	    c = line_buffer[column];
	    switch (c) {
	        case 't': case 'T':		/* #T(rue) */
	        case 'f': case 'F':		/* #F(alse) */
	        case 'v': case 'V':		/* #V(oid) */
		case 'e': case 'E':		/* #E(of) */
		    token_row[t].type = Keyword;
		    while (is_letter(line_buffer[column])) column++;
		    break;
		case '/': case '\\':
		    token_row[t].type = Character_Name;
		    column++;
		    if (is_letter(line_buffer[column])) {
			do column++; while (is_letter(line_buffer[column]));
		    } else {
		        column++;
		    }
		    break;
		case '\'':
		    column++;
		    /*FALLTHROUGH*/
		default:
		    token_row[t].type = Special;
		    break;
	    }
	    token_row[t].length = column - token_row[t].start;
	    state = Other_Token;
	} else {
	    /* plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


static char *scm_keywords[] = {
    "and",
    "begin",
    "case",
    "cond",
    "define",
    "define-syntax",
    "do",
    "do*",
    "else",
    "if",
    "lambda",
    "let",
    "let-syntax",
    "let*",
    "letrec",
    "or",
    "set!",
    (char *)0
};

struct Language scm = {
    "Scheme",
    ".scm.sc",
    "([{",				/* left brackets */
    ")]}",				/* right brackets */
    "~!@$%^&*_-+=<>?/.0123456789",	/* other identifier chars */
    "'`,",				/* special characters */
    '\\',
    '\0',				/* no user-defined operators */
    Not_Case_Sensitive,
    scm_keywords,
    tokenise
};

