/*  File   : sgml.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for SGML (really HTML)
    Beware : This is heckuva crude; it should be ok for HTML.
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 sgml.c	1.14";
#endif/*lint*/

#include "many2h.h"
#include "language.h"

#define ZERO	    Term_Ender		/* after <x> */
#define ONE	    Other_Token		/* after <x  */
#define TWO	    Prime         	/* after <x y= */

#define cdata	    (terminator[0] == ']')

static int net_ct = 0;			/* how many <x/ are open? */


static void tokenise(void) {
    int column, newcol;
    int t;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    if (state == String_Open) {
	newcol = terminator_column(column, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = cdata ? Special : String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = cdata ? Keyword : String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = cdata ? ZERO : ONE;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	    t = 0;
	}
    } else
    if (state == Comment_Open) {
	newcol = terminator_column(column, EOF);
	token_row[0].type   = Comment_Element;
	token_row[0].start  = column;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = Comment_Closer;
	token_row[1].start  = newcol;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = old_state;
	t = 2;
    } else {
	t = 0;    
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (c == '&') {
	    /* This may be the beginning of an entity */
	    for (;;) {
		if (line_buffer[column+1] == '#') {
		    if (dncase[line_buffer[column+2]] == 'x') {
			/* &#x{HEX}*[;]? */
			column += 3;  /* &#x */
			while (is_a_letter_or_digit(line_buffer[column]))
			    column++;
		    } else
		    if (is_a_digit(line_buffer[column+2])) {
			/* &#{DEC}+[;] */
			column += 3;  /* &#n */
			while (is_a_digit(line_buffer[column])) column++;
		    } else {
			token_row[t].length = 1;
			column++;
			break;
		    }
		} else
		if (is_a_letter(line_buffer[column+1])) {
		    /* &{LET}{LETDIG}*[;] */			
		    while (is_a_letter_or_digit(line_buffer[++column])) {}
		} else {
		    token_row[t].length = 1;
		    token_row[t].type = Special;
		    column++;
		    break;
		}
		if (line_buffer[column] == ';') column++;
		token_row[t].length = column - token_row[t].start;
		token_row[t].type = Identifier;
		break;
	    }
	} else
	if (c == '/') {
	    /* There are four cases:
		<xml empty element/>
		<net enabling tag/    .... /
		/                     closing a <x/
		/ as plain text
	       A fifth case is an empty end-tag </>,
	       which is handled in the '<' block.
	    */
	    if (line_buffer[column+1] == '>' && state != ZERO) {
		token_row[t].length = 2;
		token_row[t].type = Keyword;
		column += 2;
		state = ZERO;
	    } else {
		token_row[t].length = 1;
		column++;
		if (state != ZERO) {
		    /* <net enabling tag/ */
		    token_row[t].type = Keyword;
		    state = ZERO;
		    net_ct++;
		} else
		if (net_ct > 0) {
		    /* / that ends a <x/ element */
		    token_row[t].type = Keyword;
		    net_ct--;
		} else {
		    /* / that is plain text */
		    token_row[t].type = Special;
		}
	    }
	} else
	if (c == '>' && state != ZERO) {
	    token_row[t].length  = 1;
	    token_row[t].type = Keyword;
	    column++;
	    state = ZERO;
	} else
	if (c == '<') {
/* CHECK for <!CDATA[...]]> here */
	    column++;
	    c = line_buffer[column];
	    if (c == '>') {
		token_row[t].length = 1;
		token_row[t].type = Keyword;
		state = ONE;
	    } else
	    if (is_a_letter(c) || c == '?' || c == '[' ||
	        c == '/' || c == '!') {
		if (is_a_letter(c) || is_a_letter(line_buffer[++column]))
		    while (is_a_letter_or_digit(line_buffer[++column])) {}
		token_row[t].length = column - token_row[t].start;
		token_row[t].type = Keyword;
		state = ONE;
	    } else {
	        token_row[t].length = 1;
	        token_row[t].type = Special;
	    }
	} else
	if (c == '-' && state != ZERO && line_buffer[column+1] == c) {
	    static unsigned char const ender[3] = {'-', '-', '\0'};

	    token_row[t].type = Comment_Opener;
	    token_row[t].length = 2;
	    column += 2;
	    while (line_buffer[column] == ' ') column++;
	    terminator = ender;
	    termlength = 2;
	    newcol = terminator_column(column, EOF);
	    if (line_buffer[column] != '\0' && newcol != column) {
		t++;
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		if (newcol >= 0) {
		    while (newcol > 0 && line_buffer[newcol-1] == ' ')
			newcol--;
		    token_row[t].length = newcol - column;
		    newcol = terminator_column(newcol, EOF);
		} else {
		    token_row[t].length = line_length - column;
		}
	    }
	    if (newcol >= column) {
		t++;
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[newcol];
		token_row[t].start  = newcol;
		token_row[t].length = termlength;
		column = newcol+termlength;
	    } else {
		old_state = state;
		state = Comment_Open;
		column = line_length;
	    }
	} else
	if (c == '=') {
	    token_row[t].length = 1;
	    column++;
	    if (state == ZERO) {
		token_row[t].type = Special;
	    } else {
		token_row[t].type = Tick;
		state = TWO;
	    }
	} else
	if ((c == '"' || c == '\'') && state != ZERO) {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = ONE;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = ONE;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (is_a_letter_or_digit(c)) {
	    while (is_a_letter_or_digit(line_buffer[++column])) {}
	    token_row[t].length = column - token_row[t].start;
	    switch (state) {
		case TWO:
		    state = ONE;
		    token_row[t].type = String_Element;
		    break;
		case ONE:
		    token_row[t].type = Attribute;
		    break;
		default:
		    token_row[t].type = Special;
		    break;
	    }
	} else {
	    token_row[t].length = 1;
	    token_row[t].type = Special;
	    column++;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


static char *keywords[] = {
    (char*)0
};

struct Language sgml = {
    "HTML/XML/SGML",
    ".xml.html.htm.sgml.sgm.dtd.ent.xsl.rdf",
    "([{",
    ")]}",
    "-._:",
    "",
    '\0',
    '\0',
    Not_Case_Sensitive,
    keywords,
    tokenise
};

