/*  File   : sml.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for SML
    BEWARE : This should be just about right, except that SML comments
             (*(*NEST*)*) and this doesn't handle that yet.
*/

#include "many2h.h"
#include "language.h"


static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    t = 0;
    if (state == String_Open) {
        if (line_buffer[column] == escaper) column++;
	newcol = terminator_column(column, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = Term_Ender;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	}
    } else
    if (state == Comment_Open) {
	newcol = terminator_column(column, EOF);
	token_row[0].type   = Comment_Element;
	token_row[0].start  = column;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = Comment_Closer;
	token_row[1].start  = newcol;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = old_state;
	t = 2;
    }

    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter_or_underscore(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].length = column - token_row[t].start;
	    k = is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = is_a_letter(c) ? Identifier : Attribute;
		state = Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {

		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    token_row[t].type   = Number;
	    newcol = column;
	    do column++;
		while (is_a_digit_or_underscore(line_buffer[column]));
	    c = line_buffer[column];
	    if (c == '.') {
		do column++;
		    while (is_a_digit_or_underscore(line_buffer[column]));
		if (dncase[line_buffer[column]] == 'e'
		 && !is_a_letter(line_buffer[column+1])
                ) {
		    column++;
		    c = line_buffer[column];
		    if (c == '+' || c == '-') column++;
		    while (is_a_digit_or_underscore(line_buffer[column]))
			column++;
		}
	    }
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
	} else
	if (c == '(' && line_buffer[column+1] == '*') {
	    static unsigned char const ender[3] = {'*', ')', '\0'};

	    token_row[t].type = Comment_Opener;
	    token_row[t].length = 2;
	    column += 2;
	    while (line_buffer[column] == ' ') column++;
	    terminator = ender;
	    termlength = 2;
	    newcol = terminator_column(column, EOF);
	    if (line_buffer[column] != '\0' && newcol != column) {
		t++;
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		if (newcol >= 0) {
		    while (newcol > 0 && line_buffer[newcol-1] == ' ')
			newcol--;
		    token_row[t].length = newcol - column;
		    newcol = terminator_column(newcol, EOF);
		} else {
		    token_row[t].length = line_length - column;
		}
	    }
	    if (newcol >= column) {
		t++;
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[newcol];
		token_row[t].start  = newcol;
		token_row[t].length = termlength;
		column = newcol+termlength;
	    } else {
		old_state = state;
		state = Comment_Open;
		column = line_length;
	    }
	} else
	if (c == '"') {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == ',' || c == ';' || c == '|') {
	    token_row[t].type    = Comma;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_special_character(c)) {
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	} else {
	    /* @ $ plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


static char *sml_keywords[] = {
    "abs",			/* technically not reserved */
    "abstype",
    "and",
    "andalso",
    "as",
    "case",
    "div",			/* technically not reserved */
    "do",
    "datatype",
    "else",
    "end",
    "eqtype",
    "exception",
    "fn",
    "fun",
    "functor",
    "handle",
    "if",
    "in",
    "include",
    "infix",
    "infixr",
    "let",
    "local",
    "mod",			/* technically not reserved */
    "nonfix",
    "o",			/* technically not reserved */
    "of",
    "op",
    "open",
    "orelse",
    "raise",
    "rec",
    "sharing",
    "sig",
    "signature",
    "struct",
    "structure",
    "then",
    "type",
    "use",			/* technically not reserved */
    "val",
    "with",
    "withtype",
    "while",
    (char *)0
};

struct Language sml = {
    "SML",
    ".sml.sml-sig.sig",
    "([{",
    ")]}",
    "_'",
    "~!%$#`^&@*-+=|\\<>?:/.",
    '\\',
    '\0',	/* user-defined operators are done another way */
    Is_Case_Sensitive,
    sml_keywords,
    tokenise
};

