/*  File   : sql.c
    Author : Richard A. O'Keefe
    Updated: 2011/02/02
    Purpose: Tokeniser for SQL
*/

#ifndef	lint
static	char SCCSid[] = "@(#)11/02/02 sql.c	1.5";
#endif/*lint*/

#include "many2h.h"
#include "language.h"


/*  We generally try to keep track of whether we are expecting
    an operator or an operand, so that we can tell the difference
    between Id'Attr and 'x' in Ada and the difference between
    x/y and x ~ /y/ in AWK.  There is no particular difficulty in
    doing this in SQL, but there is also nothing to gain.  This
    file was copied from the C file, so it doesn't bother with
    the error reports, but does still keep track.
*/
static void tokenise(void) {
    int column, newcol;
    int t;
    int k;
    unsigned char c;

    for (column = 0; line_buffer[column] == ' '; column++) {}
    if (state == String_Open) {
	newcol = terminator_column(column, escaper);
	if (newcol >= -1) {
	    token_row[0].type   = String_Element;
	    token_row[0].start  = column;
	    token_row[0].text   = &line_buffer[column];
	    if (newcol < 0) {
		token_row[0].length = line_length-column;
		token_row[1].type   = End_Of_Line;
		return;
	    }
	    token_row[0].length = newcol-column;
	    token_row[1].type   = String_Closer;
	    token_row[1].start  = newcol;
	    token_row[1].text   = &line_buffer[newcol];
	    token_row[1].length = termlength;
	    column = newcol + token_row[1].length;
	    while (line_buffer[column] == ' ') column++;
	    state = Term_Ender;
	    t = 2;
	} else {
	    where_am_i(0);	    
	    fprintf(stderr, "Missing closing %s\n", (char const *)terminator);
	    t = 0;
	}
    } else
    if (state == Comment_Open) {
	newcol = terminator_column(column, EOF);
	token_row[0].type   = Comment_Element;
	token_row[0].start  = column;
	token_row[0].text   = &line_buffer[column];
	if (newcol < 0) {
	    token_row[0].length = line_length-column;
	    token_row[1].type = End_Of_Line;
	    return;
	}
	token_row[0].length = newcol-column;
	token_row[1].type   = Comment_Closer;
	token_row[1].start  = newcol;
	token_row[1].text   = &line_buffer[newcol];
	token_row[1].length = termlength;
	column = newcol + token_row[1].length;
	while (line_buffer[column] == ' ') column++;
	state = old_state;
	t = 2;
    } else {
	t = 0;    
    }
    while ((c = line_buffer[column]) != '\0') {
	token_row[t].start = column;
	token_row[t].text  = &line_buffer[column];

	if (is_a_letter_or_underscore(c)) {
	    do column++; while (is_a_letter_or_digit(line_buffer[column]));
	    token_row[t].length = column - token_row[t].start;
	    /* I really want to handle <word>' as a fat quote */
	    /* but for now it suffices not to treat it as a keyword */
	    k = line_buffer[column] == '\'' ? 0 : is_keyword(&token_row[t]);
	    if (k == 0) {
		token_row[t].type = Identifier;
		state = t == 2 && token_row[0].text[0] == '#'
		      ? Other_Token : Term_Ender;
	    } else
	    if (k > 0) {
		token_row[t].type = Keyword;
		state = Other_Token;
	    } else {
		token_row[t].type = Comment_Opener;
		t++;
		while (line_buffer[column] == ' ') column++;
		if (line_buffer[column] != '\0') {
		    token_row[t].type   = Comment_Element;
		    token_row[t].text   = &line_buffer[column];
		    token_row[t].start  = column;
		    token_row[t].length = line_length - column;
		    t++;
		}
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = 0;
		column = line_length;
	    }
	} else
	if (is_a_digit(c) || (c == '.' && is_a_digit(line_buffer[column+1]))) {
	    newcol = column;
	    do column++;
		while (is_a_digit_or_underscore(line_buffer[column]));
	    c = line_buffer[column];
	    if (dncase[c] == 'x' && line_buffer[newcol] == '0'
             && column - newcol == 1
	    ) {
		do column++;
		    while (is_a_letter_or_digit(line_buffer[column]));
	    } else
	    if (c == '.') {
		do column++;
		    while (is_a_digit_or_underscore(line_buffer[column]));
		if (dncase[line_buffer[column]] == 'e'
		 && !is_a_letter(line_buffer[column+1])
                ) {
		    column++;
		    c = line_buffer[column];
		    if (c == '+' || c == '-') column++;
		    while (is_a_digit_or_underscore(line_buffer[column]))
			column++;
		}
	    }
	    while (is_a_letter(line_buffer[column])) column++;
	    token_row[t].type   = Number;
	    token_row[t].length = column - token_row[t].start;
	    state = Term_Ender;
	} else
	if (c == '-' && line_buffer[column+1] == c) {
	    token_row[t].type = Comment_Opener;
	    column += 2;
	    while (line_buffer[column] == c) column++;
	    token_row[t].length = column - token_row[t].start;
	    t++;
	    while (line_buffer[column] == ' ') column++;
	    if (line_buffer[column] != '\0') {
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		token_row[t].length = line_length - column;
		t++;
	    }
	    token_row[t].type   = Comment_Closer;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    token_row[t].length = 0;
	    column = line_length;
	} else
	if (c == '/' && line_buffer[column+1] == '*') {
	    static unsigned char const ender[3] = {'*', '/', '\0'};

	    token_row[t].type = Comment_Opener;
	    token_row[t].length = 2;
	    column += 2;
	    while (line_buffer[column] == ' ') column++;
	    terminator = ender;
	    termlength = 2;
	    newcol = terminator_column(column, EOF);
	    if (line_buffer[column] != '\0' && newcol != column) {
		t++;
		token_row[t].type   = Comment_Element;
		token_row[t].text   = &line_buffer[column];
		token_row[t].start  = column;
		if (newcol >= 0) {
		    while (newcol > 0 && line_buffer[newcol-1] == ' ')
			newcol--;
		    token_row[t].length = newcol - column;
		    newcol = terminator_column(newcol, EOF);
		} else {
		    token_row[t].length = line_length - column;
		}
	    }
	    if (newcol >= column) {
		t++;
		token_row[t].type   = Comment_Closer;
		token_row[t].text   = &line_buffer[newcol];
		token_row[t].start  = newcol;
		token_row[t].length = termlength;
		column = newcol+termlength;
	    } else {
		old_state = state;
		state = Comment_Open;
		column = line_length;
	    }
	} else
	if (c == '`') {
	    lq_count++;
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (c == '\'' && lq_count > 0) {
	    lq_count--;
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == '"' || c == '\'') {
	    static unsigned char ender[2];

	    ender[0] = c, ender[1] = '\0';
	    token_row[t].type   = String_Opener;
	    token_row[t].length = 1;
	    t++;
	    column++;
	    terminator = ender;
	    termlength = 1;
	    token_row[t].type   = String_Element;
	    token_row[t].text   = &line_buffer[column];
	    token_row[t].start  = column;
	    newcol = terminator_column(column, escaper);
	    if (newcol >= 0) {
	        token_row[t].length = newcol - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = &line_buffer[newcol];
	        token_row[t].start  = newcol;
	        token_row[t].length = termlength;
	        column = newcol + termlength;
		state = Term_Ender;
	    } else
	    if (newcol < -1) {
		where_am_i(column);
		fprintf(stderr, "Missing closing %c\n", c);
	        token_row[t].length = line_length - column;
	        t++;
	        token_row[t].type   = String_Closer;
	        token_row[t].text   = token_row[t-2].text;
	        token_row[t].start  = line_length;
	        token_row[t].length = token_row[t-2].length;
	        column = line_length;
		state = Term_Ender;
	    } else {
		token_row[t].length = line_length - column;
		t++;
		token_row[t].type   = End_Of_Line;
		column = line_length;
		state = String_Open;
	    }
	} else
	if (is_a_left_bracket(c)) {
	    token_row[t].type    = Left_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_right_bracket(c)) {
	    token_row[t].type    = Right_Bracket;
	    token_row[t].length  = 1;
	    column++;
	    state = Term_Ender;
	} else
	if (c == ',') {
	    token_row[t].type    = Comma;
	    token_row[t].length  = 1;
	    column++;
	    state = Other_Token;
	} else
	if (is_a_special_character(c)) {
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	} else {
	    /* @ $ plus Latin-1 specials */
	    where_am_i(column);
	    fprintf(stderr, "'%c' [%02X] cannot begin a token\n", c, c);
	    token_row[t].type   = Special;
	    token_row[t].length = 1;
	    column++;
	    state = Other_Token;
	}
	while (line_buffer[column] == ' ') column++;
	t++;
    }
    token_row[t].type   = End_Of_Line;
    token_row[t].text   = &line_buffer[column];
    token_row[t].start  = column;
    token_row[t].length = 0;
}


/*  This table is supposed to include all the reserved keywords in
    SQL1, SQL2, and SQL3.  It is NOT supposed to include any
    non-reserved keywords.  It is also supposed to include most
    vendor-specific reserved words.  I would be delighted to show
    which vendors use what extra keywords, but the table I took
    this from didn't show that; "v" = "some major vendor".
*/
static char *keywords[] = {
    "abort",		/*v*/
    "absolute",
    "accept",		/*v*/
    "action",
    "add",
    "admin",
    "after",
    "aggregate",
    "alias",
    "all",
    "allocate",
    "alter",
    "analyze",		/*v*/
    "and",
    "any",
    "archivelog",	/*v*/
    "are",
    "array",
    "as",
    "asc",
    "assertion",
    "assign",		/*v*/
    "asynch",		/*v*/
    "at",
    "attributes",	/*v*/
    "audit",		/*v*/
    "authorization",
    "backup",		/*v*/
    "before",
    "begin",
    "binary",
    "binary_integer",	/*v*/
    "bit",
    "blob",
    "body",		/*v*/
    "boolean",
    "both",
    "breadth",
    "by",
    "cache",		/*v*/
    "call",
    "cascade",
    "cascaded",
    "case",
    "cast",
    "catalog",
    "char",
    "char_base",	/*v*/
    "character",
    "check",
    "class",
    "clob",
    "close",
    "cluster",		/*v*/
    "clusters",		/*v*/
    "colauth",		/*v*/
    "collate",
    "collation",
    "column",
    "columns",		/*v*/
    "commit",
    "completion",
    "compress",		/*v*/
    "condition",
    "connect",
    "connection",
    "constant",		/*v*/
    "constraint",
    "constraints",
    "constructor",
    "contains",
    "continue",
    "corresponding",
    "crash",		/*v*/
    "create",
    "cross",
    "cube",
    "current",
    "current_date",
    "current_path",
    "current_role",
    "current_time",
    "current_timestamp",
    "current_user",
    "cursor",
    "curval",		/*v*/
    "cycle",
    "data_base",	/*v*/
    "data",
    "database",		/*v*/
    "datalink",
    "date",
    "day",
    "dba",		/*v*/
    "deallocate",
    "debugoff",		/*v*/
    "debugon",		/*v*/
    "dec",
    "decimal",
    "declare",
    "default",
    "deferrable",
    "deferred",
    "definition",	/*v*/
    "delay",		/*v*/
    "delete",
    "delta",		/*v*/
    "depth",
    "deref",
    "desc",
    "describe",
    "descriptor",
    "destroy",
    "destructor",
    "deterministic",
    "diagnostics",
    "dictionary",
    "digits",		/*v*/
    "disconnect",
    "displacement",	/*v*/
    "dispose",		/*v*/
    "distinct",
    "do",
    "domain",
    "double",
    "drop",
    "dynamic",
    "each",
    "element",		/*v*/
    "else",
    "elseif",
    "end",
    "end-exec",
    "entry",		/*v*/
    "equals",
    "escape",
    "every",
    "exec",
    "except",
    "exception",
    "exception_init",	/*v*/
    "execute",
    "exit",
    "expand",
    "expanding",
    "external",
    "factor",		/*v*/
    "false",
    "fetch",
    "first",
    "float",
    "for",
    "foreign",
    "form",		/*v*/
    "found",
    "from",
    "free",
    "freelists",	/*v*/
    "full",
    "function",
    "general",
    "generic",		/*v*/
    "get",
    "global",
    "go",
    "goto",
    "grant",
    "group",
    "grouping",
    "handler",
    "having",
    "hash",
    "host",
    "hour",
    "identified",	/*v*/
    "identity",
    "if",
    "ignore",
    "immediate",
    "in",
    "include",		/*v*/
    "index",		/*v*/
    "indexes",		/*v*/
    "indicator",
    "infile",		/*v*/
    "initialize",
    "initially",
    "inner",
    "inout",
    "input",
    "insert",
    "instead",		/*v*/
    "instance",		/*v*/
    "int",
    "integer",
    "intersect",
    "interval",
    "into",
    "is",
    "isolation",
    "iterate",
    "join",
    "key",
    "language",
    "large",
    "last",
    "lateral",
    "leading",
    "leave",
    "left",
    "less",
    "level",
    "like",
    "limit",
    "limited",		/*v*/
    "list",		/*v*/
    "local",
    "localtime",
    "localtimestamp",
    "locator",
    "loop",
    "match",
    "maxextents",	/*v*/
    "meets",
    "minus",		/*v*/
    "minute",
    "mlslabel",		/*v*/
    "mode",		/*v*/
    "modifies",
    "modify",
    "moduole",
    "month",
    "names",
    "national",
    "natural",
    "nchar",
    "nclob",
    "new",
    "new_table",	/*v*/
    "next",
    "nextval",		/*v*/
    "no",
    "nocompress",	/*v*/
    "none",
    "normalize",
    "not",
    "null",
    "number",		/*v*/
    "number_base",	/*v*/
    "numeric",
    "object",
    "of",
    "off",
    "oid",		/*v*/
    "old",
    "old_table",	/*v*/
    "on",
    "only",
    "open",
    "operation",
    "operator",		/*v*/
    "operators",	/*v*/
    "option",
    "or",
    "order",
    "ordinality",
    "others",		/*v*/
    "out",
    "outer",
    "output",
    "package",		/*v*/
    "pad",
    "parameter",
    "parameters",
    "partial",
    "partition",	/*v*/
    "path",
    "pctfree",		/*v*/
    "pendant",		/*v*/
    "period",
    "positive",		/*v*/
    "postfix",
    "pragma",		/*v*/
    "precedes",
    "precision",
    "prefix",
    "preorder",
    "preordered",	/*v*/
    "prepare",
    "preserve",
    "primary",
    "prior",
    "private",		/*v*/
    "privileges",
    "procedure",
    "protected",	/*v*/
    "public",
    "raise",		/*v*/
    "range",		/*v*/
    "raw",		/*v*/
    "read",
    "reads",
    "real",
    "record",		/*v*/
    "recursive",
    "redo",
    "ref",
    "references",
    "referencing",
    "relative",
    "release",		/*v*/
    "rename",		/*v*/
    "repeat",
    "replace",		/*v*/
    "resignal",
    "resource",		/*v*/
    "restrict",
    "result",
    "return",
    "returns",
    "reuse",		/*v*/
    "reverse",		/*v*/
    "revoke",
    "right",
    "role",
    "rollback",
    "rollup",
    "routine",
    "row",
    "rowid",		/*v*/
    "rowlabel",		/*v*/
    "rownum",		/*v*/
    "rows",
    "rowtype",		/*v*/
    "run",		/*v*/
    "savepoint",
    "schema",
    "scroll",
    "search",
    "second",
    "section",
    "select",
    "separate",		/*v*/
    "sequence",
    "session",
    "session_user",
    "set",
    "sets",
    "signal",
    "size",
    "smallint",
    "some",
    "space",
    "specific",
    "specifictype",
    "sql",
    "sqlca",		/*v*/
    "sqlcode",		/*v*/
    "sqlerrm",		/*v*/
    "sqlexception",
    "sqlstate",
    "sqlwarning",
    "start",
    "state",
    "statement",	/*v*/
    "static",
    "stddev",		/*v*/
    "structure",
    "subtype",		/*v*/
    "succeeds",
    "symbol",		/*v*/
    "system_user",
    "tabauth",		/*v*/
    "table",
    "tables",		/*v*/
    "task",		/*v*/
    "temporary",
    "term",		/*v*/
    "terminate",
    "test",		/*v*/
    "than",
    "then",
    "there",		/*v*/
    "time",
    "timestamp",
    "timezone_hour",
    "timezone_minute",
    "to",
    "trailing",
    "transaction",
    "translation",
    "treat",
    "trigger",
    "true",
    "tuple",		/*v*/
    "under",
    "undo",
    "union",
    "unique",
    "unknown",
    "until",
    "update",
    "usage",
    "use",		/*v*/
    "user",
    "using",
    "value",
    "values",
    "varchar",
    "varchar2",		/*v*/
    "variable",
    "variance",		/*v*/
    "varying",
    "view",
    "views",		/*v*/
    "virtual",		/*v*/
    "visible",		/*v*/
    "wait",		/*v*/
    "when",
    "whenever",
    "where",
    "while",
    "with",
    "without",
    "work",
    "write",
    "xor",		/*v*/
    "year",    
    "zone",
    (char *)0
};

struct Language sql = {
    "SQL",
    ".sql",
    "([{",
    ")]}",
    "_",
    "%&*/+-.:;<>?|=^",
    '\0',
    '\0',
    Not_Case_Sensitive,
    keywords,
    tokenise
};


